#![allow(dead_code)]
pub mod brain {
    tonic::include_proto!("brain");
}

use anyhow::Result;
use brain::{
    brain_daemon_service_client::BrainDaemonServiceClient, DeleteVmReq, NewVmReq, NewVmResp,
    NodePubkey, NodeResourceReq, RegisterNodeReq, UpdateVmReq, UpdateVmResp,
};
use lazy_static::lazy_static;
use log::{debug, error, info, warn};
use rand::{distributions::Alphanumeric, thread_rng, Rng};
use tokio::time::{sleep, Duration};
use tokio::{sync::mpsc::Receiver, sync::mpsc::Sender, task::JoinSet};
use tokio_stream::wrappers::ReceiverStream;
use tokio_stream::StreamExt;
use tonic::transport::Channel;

lazy_static! {
    static ref SECURE_PUBLIC_KEY: String = generate_random_string();
}

fn generate_random_string() -> String {
    let rng = thread_rng();
    rng.sample_iter(&Alphanumeric)
        .take(16) // Adjust the length as needed
        .map(char::from)
        .collect()
}

async fn listen_for_new_vm_reqs(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<NewVmReq>,
) -> Result<()> {
    debug!("starting listen_for_new_vm_reqs");
    let node_pubkey = SECURE_PUBLIC_KEY.clone();
    let mut grpc_stream = client
        .get_new_vm_reqs(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(stream_update) = grpc_stream.next().await {
        match stream_update {
            Ok(req) => {
                info!("Received new vm request: {req:?}");
                let _ = tx.send(req).await;
            }
            Err(e) => {
                warn!("Brain disconnected from listen_for_new_vm_reqs: {e}");
            }
        }
    }
    debug!("listen_for_new_vm_reqs is about to exit");
    Ok(())
}

async fn send_newvm_resp(
    mut client: BrainDaemonServiceClient<Channel>,
    rx: Receiver<NewVmResp>,
) -> Result<()> {
    debug!("starting send_newvm_resp stream");
    let rx_stream = ReceiverStream::new(rx);
    client.send_new_vm_resp(rx_stream).await?;
    debug!("send_newvm_resp is about to exit");
    Ok(())
}

async fn send_node_resources(
    mut client: BrainDaemonServiceClient<Channel>,
    rx: Receiver<NodeResourceReq>,
) -> Result<()> {
    debug!("starting send_newvm_resp stream");
    let rx_stream = ReceiverStream::new(rx);
    client.send_node_resources(rx_stream).await?;
    debug!("send_newvm_resp is about to exit");
    Ok(())
}

async fn register_node(mut client: BrainDaemonServiceClient<Channel>) {
    debug!("Starting node registration...");
    let req = RegisterNodeReq {
        node_pubkey: SECURE_PUBLIC_KEY.clone(),
        owner_pubkey: "IamTheOwnerOf".to_string() + &SECURE_PUBLIC_KEY,
    };
    match client.register_node(req).await {
        Ok(_) => info!(
            "Registered as 10.0.10.1 from Bruma/Cyrodiil with ID {}",
            SECURE_PUBLIC_KEY.clone()
        ),
        Err(e) => error!("Could not register node data: {e:?}"),
    };
}

async fn listen_for_deleted_vms(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<DeleteVmReq>,
) -> Result<()> {
    debug!("starting listen_for_new_vm_reqs");
    let node_pubkey = SECURE_PUBLIC_KEY.clone();
    let mut grpc_stream = client
        .get_delete_vm_req(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(stream_update) = grpc_stream.next().await {
        match stream_update {
            Ok(req) => {
                info!("Received delete vm request: {req:?}");
                let _ = tx.send(req).await;
            }
            Err(e) => {
                warn!("Brain disconnected from listen_for_deleted_vms: {e}");
            }
        }
    }
    debug!("listen_for_new_vm_reqs is about to exit");
    Ok(())
}

async fn listen_for_update_vm_reqs(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<UpdateVmReq>,
) -> Result<()> {
    debug!("starting listen_for_update_vm_reqs");
    let node_pubkey = SECURE_PUBLIC_KEY.clone();
    let mut grpc_stream = client
        .get_update_vm_req(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(stream_update) = grpc_stream.next().await {
        match stream_update {
            Ok(req) => {
                info!("Received update vm request: {req:?}");
                let _ = tx.send(req).await;
            }
            Err(e) => {
                warn!("Brain disconnected from listen_for_update_vm_reqs: {e}");
            }
        }
    }
    debug!("listen_for_update_vm_reqs is about to exit");
    Ok(())
}

async fn handle_update_vm_requests(
    mut req: Receiver<UpdateVmReq>,
    resp_chan: Sender<UpdateVmResp>,
) {
    info!("Started to handle update vm requests.");
    while let Some(update_vm) = req.recv().await {
        let update_vm_resp = UpdateVmResp {
            uuid: update_vm.uuid,
            error: "".to_string(),
        };
        info!("Sending UpdateVmResp: {update_vm_resp:?}");
        let _ = resp_chan.send(update_vm_resp).await;
    }
    warn!("update vm request handler is ending");
}

async fn send_updatevm_resp(
    mut client: BrainDaemonServiceClient<Channel>,
    rx: Receiver<UpdateVmResp>,
) -> Result<()> {
    debug!("starting send_updatevm_resp stream");
    let rx_stream = ReceiverStream::new(rx);
    client.send_update_vm_resp(rx_stream).await?;
    debug!("send_updatevm_resp is about to exit");
    Ok(())
}

async fn handle_vm_requests(
    mut req: Receiver<NewVmReq>,
    resp: Sender<NewVmResp>,
    resource_tx: Sender<NodeResourceReq>,
) {
    info!("Started to handle vm requests. 1 out of 5 requests will return error.");
    let mut i = 0;
    while let Some(new_vm) = req.recv().await {
        let exposed_ports = match new_vm.public_ipv4 {
            true => Vec::new(),
            false => vec![20321, 20415, 25912],
        };
        let public_ipv4 = match new_vm.public_ipv4 {
            true => "10.0.100.5".to_string(),
            false => String::new(),
        };
        let public_ipv6 = match new_vm.public_ipv6 {
            true => "2a02:2f2d:d301:3100:afe8:a85e:54a0:dd28".to_string(),
            false => String::new(),
        };
        if i != 3 {
            let confirmation = NewVmResp {
                uuid: new_vm.uuid,
                exposed_ports,
                public_ipv4,
                public_ipv6,
                error: String::new(),
            };
            info!("Sending NewVmConfirmation: {confirmation:?}");
            let _ = resp.send(confirmation).await;
            info!("Sending NodeResourceReq");
        } else {
            let confirmation = NewVmResp {
                uuid: new_vm.uuid,
                exposed_ports: Vec::new(),
                public_ipv4: String::new(),
                public_ipv6: String::new(),
                error: "No.".to_string(),
            };
            info!("Sending error for NewVmConfirmation: {confirmation:?}");
            let _ = resp.send(confirmation).await;
            info!("Sending updated about used resources.");
            let _ = resource_tx
                .send(NodeResourceReq {
                    node_pubkey: SECURE_PUBLIC_KEY.clone(),
                    avail_ipv4: 5,
                    avail_ipv6: 5,
                    avail_memory_mb: 5000,
                    avail_vcpus: 32,
                    avail_storage_gb: 100,
                    avail_ports: 5000,
                    max_ports_per_vm: 5,
                })
                .await;
        }
        i += 1;
        if i == 5 {
            i = 0;
        }
    }
    warn!("vm request handler is ending");
}

async fn connect_and_run() -> Result<()> {
    let client = BrainDaemonServiceClient::connect("http://[::1]:31337").await?;
    let mut streaming_tasks = JoinSet::new();

    register_node(client.clone()).await;

    let newvm_client = client.clone();
    let (tx, newvm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_new_vm_reqs(newvm_client, tx));

    let confirm_client = client.clone();
    let resource_client = client.clone();
    let (resource_tx, resource_rx) = tokio::sync::mpsc::channel(6);
    let _ = resource_tx
        .send(NodeResourceReq {
            node_pubkey: SECURE_PUBLIC_KEY.clone(),
            avail_ipv4: 10,
            avail_ipv6: 10,
            avail_memory_mb: 2500,
            avail_vcpus: 60,
            avail_storage_gb: 200,
            avail_ports: 5002,
            max_ports_per_vm: 5,
        })
        .await;
    let (confirm_tx, confirm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(send_newvm_resp(confirm_client, confirm_rx));
    streaming_tasks.spawn(send_node_resources(resource_client, resource_rx));
    tokio::spawn(async move {
        handle_vm_requests(newvm_rx, confirm_tx, resource_tx).await;
    });

    let updatevm_client = client.clone();
    let (tx, updatevm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_update_vm_reqs(updatevm_client, tx));

    let resp_client = client.clone();
    let (resp_tx, rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(send_updatevm_resp(resp_client, rx));

    tokio::spawn(async move {
        handle_update_vm_requests(updatevm_rx, resp_tx).await;
    });

    let deletevms_client = client.clone();
    let (tx, _deletevm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_deleted_vms(deletevms_client, tx));

    let task_output = streaming_tasks.join_next().await;
    warn!("One stream exited: {task_output:?}");
    Ok(())
}

async fn connection_wrapper() -> ! {
    loop {
        info!("Connecting to brain...");
        if let Err(e) = connect_and_run().await {
            error!("The connection broke: {e}");
        }
        sleep(Duration::from_secs(3)).await;
    }
}

#[tokio::main]
async fn main() {
    env_logger::builder()
        .filter_level(log::LevelFilter::Debug)
        .init();
    info!("Hello! My name is {}", SECURE_PUBLIC_KEY.clone());
    connection_wrapper().await;
}
